from .batch_filter import BatchFilter
from gunpowder.array import Array
from gunpowder.batch import Batch
from gunpowder.profiling import Timing
import numpy as np

class Stack(BatchFilter):
    '''Request several batches and stack them together, introducing a new
    dimension for each array. This is useful to create batches with several
    samples and only makes sense if there is a source of randomness upstream.

    This node stacks only arrays, not points. The resulting batch will have the
    same point sets as found in the first batch requested upstream.

    Args:

        num_repetitions (``int``):

            How many upstream batches to stack.
    '''

    def __init__(self, num_repetitions):

        self.num_repetitions = num_repetitions

    def provide(self, request):

        timing = Timing(self)
        timing.start()

        batches = [
            self.get_upstream_provider().provide(request)
            for _ in range(self.num_repetitions)
        ]

        batch = Batch()
        for key, spec in request.array_specs.items():

            data = np.stack([b[key].data for b in batches])
            batch[key] = Array(
                data,
                batches[0][key].spec.copy())

        # copy points of first batch requested
        for key, spec in request.points_specs.items():
            batch[key] = batches[0][key]

        timing.stop()
        batch.profiling_stats.add(timing)

        return batch



from .freezable import Freezable
from .graph import Graph, Node, GraphKey, GraphKeys

import logging
import warnings
import numpy as np

logger = logging.getLogger(__name__)


class Points(Graph):
    """An Alias of Graph that supports the points API

    Args:

        data (``dict``, ``int`` -> :class:`Point`):

            A dictionary of IDs mapping to :class:`Points<Point>`.

        spec (:class:`PointsSpec`):

            A spec describing the data.
    """

    def __init__(self, data, spec):
        warnings.warn("Points are depricated. Please use Graph")
        nodes = [Node(id=i, location=p.location) for i, p in data.items()]
        super().__init__(nodes, [], spec)
        self.__spec = spec
        self.freeze()

    @property
    def data(self):
        return {v.id: Point(v.location) for v in self.nodes}

    @property
    def directed(self):
        return True

    def __repr__(self):
        return "%s, %s" % (self.data, self.spec)


class Point(Freezable):
    """A point with a location, as stored in :class:`Points`.

    Args:

        location (array-like of ``float``):

            The location of this point.
    """

    def __init__(self, location):
        self.location = np.array(location, dtype=np.float32)
        self.freeze()

    def __repr__(self):
        return str(self.location)

    def copy(self):
        return Point(self.location)


PointsKey = GraphKey

PointsKeys = GraphKeys



import numpy as np

import copy

from .freezable import Freezable


class GraphSpec(Freezable):
    """Contains meta-information about a graph. This is used by
    :class:`BatchProviders<BatchProvider>` to communicate the graphs they
    offer, as well as by :class:`Graph` to describe the data they contain.

    Attributes:

        roi (:class:`Roi`):

            The region of interested represented by this graph.

        directed (``bool``):

            Whether the graph is directed or not.
    """

    def __init__(self, roi=None, directed=True, dtype=np.float32):

        self.roi = roi
        self.directed = directed
        self.dtype = dtype

        self.freeze()

    def copy(self):
        """Create a copy of this spec."""
        return copy.deepcopy(self)

    def __eq__(self, other):

        if isinstance(other, self.__class__):
            return self.__dict__ == other.__dict__
        return NotImplemented

    def __ne__(self, other):

        if isinstance(other, self.__class__):
            return not self.__eq__(other)
        return NotImplemented

    def __repr__(self):
        r = ""
        r += "ROI: " + str(self.roi)
        return r
